(**

  On the coevaluation behavior and equivalence: 
  the Coq development

  Author: Angel Zúñiga

  email: zuniga@ciencias.unam.mx

  Postdoctoral researcher

  Instituto de Ingeniería, UNAM

  Date: September 2022

  Based on the work

  Coinductive big-step operational semantics by Xavier Leroy
  https://doi.org/10.1007/11693024_5

  its extended version

  Coinductive big-step operational semantics by Xavier Leroy and Hervé Grall 
  https://doi.org/10.1016/j.ic.2007.12.004

  and its corresponding Coq development

  https://xavierleroy.org/coindsem/doc/src.tgz

  This code is distributed under the terms of the GNU Public License v2
  https://www.gnu.org/licenses/old-licenses/gpl-2.0.html 
*)


Require Import String.

Require Import Coq.Program.Equality.

(** 
    The language:
    lambda calculus extended with constants 
*)

Definition var: Set := string. (*Variable name*)

Lemma var_eq: forall (v1 v2: var), {v1 = v2} + {v1 <> v2}.
Proof.
  apply string_dec.
Defined.


(*Abstract syntax*)
Inductive term: Set :=
| Const: nat -> term
| Var: var -> term
| Lam: var -> term -> term
| App: term -> term -> term.

Fixpoint subst (x: var) (b: term) (a:term) {struct a}: term :=
  match a with
  | Const c => Const c
  | Var y => if var_eq x y then b else Var y
  | Lam y a1 => Lam y (if var_eq x y then a1 else subst x b a1)
  | App a1 a2 => App (subst x b a1) (subst x b a2)
  end.


Inductive isvalue: term -> Prop :=
| isvalue_const: forall c, isvalue (Const c)
| isvalue_fun: forall x b, isvalue (Lam x b).


(** Evaluation **)

Inductive eval: term -> term -> Prop :=
| eval_const: forall c, eval (Const c) (Const c)

| eval_fun: forall x a, eval (Lam x a) (Lam x a)

| eval_app: forall a1 x b,
    eval a1 (Lam x b) ->
    forall a2 v2,
      eval a2 v2 ->
      forall v,
    eval (subst x v2 b) v ->
    eval (App a1 a2) v.

Lemma eval_isvalue: forall a b,
    eval a b -> isvalue b.
Proof.
intros.
induction H.
constructor.
constructor.
trivial.
Qed.

(*Evaluation determinism*)
Lemma eval_deterministic:
forall a v1,
  eval a v1 ->
forall v2,    
  eval a v2 ->
  v1 = v2.
Proof.
intros.
dependent induction H.
inversion_clear H0.
trivial.
inversion_clear H0.
trivial.
inversion_clear H2.
assert((Lam x b) = (Lam x0 b0)).
apply IHeval1.
trivial.
assert(v2 = v1).
apply IHeval2.
trivial.
apply IHeval3.
congruence.
Qed.



(** Divergence **)

CoInductive evalinf: term -> Prop :=
| evalinf_app_l: forall a1,
    evalinf a1 ->
    forall a2,
    evalinf (App a1 a2)

| evalinf_app_r: forall a1 v,
    eval a1 v ->
    forall a2,
    evalinf a2 ->
    evalinf (App a1 a2)

| evalinf_app_f: forall a1 x b,
    eval a1 (Lam x b) ->
    forall a2 v,
      eval a2 v ->
      evalinf (subst x v b) ->
      evalinf (App a1 a2).


(*Evaluation and divergence are mutually exclusive*)
Lemma eval_evalinf_exclusive:
forall a v,
  eval a v ->
  evalinf a ->
  False.
Proof.
intros.
dependent induction H.
inversion H0.
inversion H0.
inversion_clear H2.
apply IHeval1.
trivial.
apply IHeval2.
trivial.
assert((Lam x b) = (Lam x0 b0)).
eapply eval_deterministic.
exact H.
trivial.
assert(v2 = v0).
eapply eval_deterministic.
exact H0.
trivial.
congruence.
Qed.


(** Coevaluation **)

CoInductive coeval: term -> term -> Prop :=
| coeval_const: forall c,
    coeval (Const c) (Const c)

| coeval_fun: forall x a,
    coeval (Lam x a) (Lam x a)

| coeval_app: forall a1 x b,
    coeval a1 (Lam x b) ->
    forall a2 v2,
    coeval a2 v2 ->
    forall v,
    coeval (subst x v2 b) v ->
    coeval (App a1 a2) v.

Lemma eval_coeval:
  forall a v,
    eval a v ->
    coeval a v.
Proof.
intros.
induction H.
constructor.
constructor.
econstructor.
exact IHeval1.
exact IHeval2.
exact IHeval3.
Qed.


Require Import Classical.

Lemma coeval_noteval_evalinf:
forall a v,
  coeval a v ->
  not(eval a v) ->
  evalinf a.
Proof.
cofix CH.

intros.
inversion H.
subst.
contradict H0.
constructor.

subst.
contradict H0.
constructor.

subst.
elim(classic (eval a1 (Lam x b))).
intros.
elim(classic (eval a2 v2)).
intros.
elim(classic (eval (subst x v2 b) v)).
intros.
elim H0.
econstructor 3.
exact H4.
exact H5.
trivial.

intros.
econstructor 3.
exact H4.
exact H5.
eapply CH.
exact H3.
trivial.

intros.
econstructor 2.
exact H4.
eapply CH.
exact H2.
trivial.

intros.
constructor 1.
eapply CH.
exact H1.
trivial.
Qed.


Lemma coeval_eval_or_evalinf:
  forall a v,
    coeval a v ->
    eval a v \/ evalinf a.
Proof.
intros.
elim(classic (eval a v)).
intros.  
left.
trivial.

intros.
right.
eapply coeval_noteval_evalinf.
exact H.
trivial.
Qed.


(** Original coevaluation unusual features **)


(*There exists terms that diverge but do not coevaluate*)

Open Scope string.

Definition delta := Lam "x" (App (Var "x") (Var "x")).
Definition omega := App delta delta.


Lemma evalinf_omega: evalinf omega.
Proof.
cofix CH.  
econstructor 3.
constructor 2.
constructor 2.
simpl.
fold delta.
fold omega.
apply CH.
Qed.

Lemma not_eval_omega: forall v, not(eval omega v).
Proof.
intros.
unfold not.
intros.
dependent induction H.
inversion H.
subst.
apply IHeval3.
simpl.
inversion H0.
subst.
fold delta.
fold omega.
trivial.
Qed.


Definition alpha := App omega (App (Const 0) (Const 0)).

Lemma evalinf_alpha: evalinf alpha.
Proof.
unfold alpha.
constructor 1.  
apply evalinf_omega.
Qed.


Lemma not_coeval_alpha: forall v, not(coeval alpha v).
Proof.
intros.
unfold not.
intros.
unfold alpha in H.
inversion_clear H.
inversion_clear H1.
inversion H.
Qed.


Lemma not_evalinf_coeval:
  not(forall a,
    evalinf a ->
    exists v, coeval a v).
Proof.
unfold not.
intros.
assert(exists v, coeval alpha v).
apply H.
apply evalinf_alpha.
destruct H0 as [v].
contradict H0.
unfold not.
apply not_coeval_alpha.
Qed.


(*There exists diverging terms which are not deterministic 
   in the sense that they coevaluate to any value v*)

Lemma coeval_omega: forall v, coeval omega v.
Proof.
cofix CH.  
econstructor 3.
constructor 2.
constructor 2.
simpl.
fold delta.
fold omega.
apply CH.
Qed.


(*There exists terms that diverge but coevaluate to only one value*)

Definition beta := App (Lam "x" (Const 0)) omega.

Lemma evalinf_beta: evalinf beta.
Proof.
unfold beta.
econstructor 2.  
constructor 2.  
apply evalinf_omega.
Qed.


Lemma coeval_beta_zero:
  coeval beta (Const 0).
Proof.
unfold beta.
econstructor 3.  
constructor 2.
apply coeval_omega with (v := (Const 0)).
simpl.
constructor 1.
Qed.


Lemma coeval_beta_only_zero:
  forall v,
    coeval beta v ->
    v = (Const 0).
Proof.
intros.
unfold beta in H.
inversion_clear H.
inversion H0.
subst.
simpl in H2.
inversion H2.
subst.
trivial.
Qed.


(** Divergence revisited **)
(* It is enhanced with an explicit singleton infinite value Div *)

Inductive dvalue: Set :=
| Div: dvalue.

CoInductive evalinfrev: term -> dvalue -> Prop :=
| evalinfrev_app_l: forall a1,
    evalinfrev a1 Div ->
    forall a2,
    evalinfrev (App a1 a2) Div

| evalinfrev_app_r: forall a1 v,
    eval a1 v ->
    forall a2,
    evalinfrev a2 Div ->
    evalinfrev (App a1 a2) Div

| evalinfrev_app_f: forall a1 x b,
    eval a1 (Lam x b) ->
    forall a2 v2,
      eval a2 v2 ->
      evalinfrev (subst x v2 b) Div ->
      evalinfrev (App a1 a2) Div.


Lemma evalinf_evalinfrev:
  forall a,
    evalinf a ->
    evalinfrev a Div.
Proof.
cofix CH.
intros.  
inversion_clear H.  
constructor 1.
apply CH.
trivial.

econstructor 2.
exact H0.
apply CH.
trivial.

econstructor 3.
exact H0.
exact H1.
apply CH.
trivial.
Qed.


Lemma evalinfrev_evalinf:
  forall a,
    evalinfrev a Div ->
    evalinf a.
Proof.
cofix CH.
intros.
inversion_clear H.
constructor 1.
apply CH.
trivial.

econstructor 2.
exact H0.
apply CH.
trivial.

econstructor 3.
exact H0.
exact H1.
apply CH.
trivial.
Qed.

(*Divergence and divergence revisited with the explicit infinite value are equivalent*)
Lemma evalinf_evalinfrev_eq:
  forall a, evalinf a <-> evalinfrev a Div.
Proof.
intros.
split.  
apply evalinf_evalinfrev.
apply evalinfrev_evalinf.
Qed.



(** Preliminary coevaluation extension **)

Inductive fdvalue: Set :=
| Fin: term -> fdvalue
| Inf: dvalue -> fdvalue.

CoInductive coevalpre: term -> fdvalue -> Prop :=
| coevalpre_const: forall c,
    coevalpre (Const c) (Fin (Const c))

| coevalpre_fun: forall x a,
    coevalpre (Lam x a) (Fin (Lam x a))

| coevalpre_app: forall a1 x b,
    coevalpre a1 (Fin (Lam x b)) ->
    forall a2 v2,
    coevalpre a2 (Fin v2) ->
    forall v,
    coevalpre (subst x v2 b) (Fin v) ->
    coevalpre (App a1 a2) (Fin v)

| coevalpre_app_f: forall a1 x b,
    coevalpre a1 (Fin (Lam x b)) ->
    forall a2 v2,
    coevalpre a2 (Fin v2) ->
    coevalpre (subst x v2 b) (Inf Div) ->
    coevalpre (App a1 a2) (Inf Div)
           
| coevalpre_app_l: forall a1,
    coevalpre a1 (Inf Div) ->
    forall a2,
    coevalpre (App a1 a2) (Inf Div)

| coevalpre_app_r: forall a1 v,
    coevalpre a1 (Fin v) ->
    forall a2,
    coevalpre a2 (Inf Div) ->
    coevalpre (App a1 a2) (Inf Div).


Lemma eval_coevalprefin:
forall a v,
eval a v ->
coevalpre a (Fin v).
Proof.
intros.
dependent induction H.
constructor.
constructor.
econstructor 3.
exact IHeval1.
exact IHeval2.
trivial.
Qed.

Lemma evalinfrev_coevalpreinf:
forall a,
  evalinfrev a Div ->
  coevalpre a (Inf Div).
Proof.
cofix CH.
intros.
inversion H.

subst.
constructor 5.
apply CH.
trivial.

subst.
econstructor 6.
apply eval_coevalprefin.
exact H0.
apply CH.
trivial.

subst.
econstructor 4.
apply eval_coevalprefin.
exact H0.
apply eval_coevalprefin.
exact H1.
apply CH.
trivial.
Qed.


Lemma evalorevalinfrev_coevalprefinorcoevalpreinf:
forall a v,
  eval a v \/ evalinfrev a Div ->
  coevalpre a (Fin v) \/ coevalpre a (Inf Div).
Proof.
intros.

destruct H.
left.
apply eval_coevalprefin.
trivial.

right.
apply evalinfrev_coevalpreinf.
trivial.
Qed.


Lemma evalorevalinfrev_coevalpre:
forall a v,
  eval a v \/ evalinfrev a Div ->
  exists vg, coevalpre a vg.
Proof.
intros.

destruct H.
exists (Fin v).
apply eval_coevalprefin.
trivial.

exists (Inf Div).
apply evalinfrev_coevalpreinf.
trivial.
Qed.


Lemma coevalprefin_noteval_evalinfrev:
  forall a v,
  coevalpre a (Fin v) ->
  not (eval a v) ->
  evalinfrev a Div.
Proof.
cofix CH.

intros.
inversion H.

subst.
elim H0.
constructor.

subst.
elim H0.
constructor.

subst.
elim (classic (eval a1 (Lam x b))).
intros.
elim (classic (eval a2  v2)).
intros.
elim (classic (eval (subst x v2 b) v)).
intros.
elim H0.
econstructor 3.
exact H1.
exact H4.
trivial.

intros.
econstructor 3.
exact H1.
exact H4.
eapply CH.
exact H5.
trivial.
intros.
econstructor 2.
exact H1.
eapply CH.
exact H3.
trivial.
intros.
constructor 1.
eapply CH.
exact H2.
trivial.
Qed.



(*These two lemmas illustrates the main problem with this preliminary extension*)
Lemma coevalpre_omega_fin: forall v, coevalpre omega (Fin v).
Proof.
cofix CH.  
econstructor 3.
constructor 2.
constructor 2.
apply CH.
Qed.

Lemma coevalpre_omega_inf: coevalpre omega (Inf Div).
Proof.
cofix CH.
econstructor 4.
constructor 2.
constructor 2.
apply CH.
Qed.


Lemma coevalprefin_evalorevalinfrev:
  forall a v,
  coevalpre a (Fin v) ->
  eval a v \/ evalinfrev a Div.
Proof.
intros.
elim (classic (eval a v)).
intros.
left.
trivial.
intros.  
right.
eapply coevalprefin_noteval_evalinfrev.
exact H.
trivial.
Qed.


Lemma coevalpreinf_evalinfrev:
forall a,
  coevalpre a (Inf Div) ->
  evalinfrev a Div.
Proof.
cofix CH.

intros.
inversion H.  

subst.
apply coevalprefin_evalorevalinfrev in H0.
destruct H0.
apply coevalprefin_evalorevalinfrev in H1.
destruct H1.
econstructor 3.
exact H0.
exact H1.
apply CH.
trivial.

econstructor 2.
exact H0.
apply CH.
apply evalinfrev_coevalpreinf.
trivial.

constructor 1.
apply CH.
apply evalinfrev_coevalpreinf.
trivial.


subst.
constructor 1.
apply CH.
trivial.

subst.
apply coevalprefin_evalorevalinfrev in H0.
destruct H0.
econstructor 2.
exact H0.
apply CH.
trivial.

constructor 1.
apply CH.
apply evalinfrev_coevalpreinf.
trivial.
Qed.

Lemma coevalpreinf_evalinfrev_eq:
  forall a, coevalpre a (Inf Div) <-> evalinfrev a Div.
Proof.
intros.
split.
apply coevalpreinf_evalinfrev.
apply evalinfrev_coevalpreinf.
Qed.


Lemma coevalprefinorcoevalpreinf_evalorevalinfrev:
  forall a v,
  coevalpre a (Fin v) \/ coevalpre a (Inf Div) ->
  eval a v \/ evalinfrev a Div.
Proof.
intros.
destruct H.
apply coevalprefin_evalorevalinfrev.
trivial.

right.
apply coevalpreinf_evalinfrev.
trivial.
Qed.


Theorem coevalprefinorcoevalpreinf_evalorevalinfrev_eq:
  forall a v,
  coevalpre a (Fin v) \/ coevalpre a (Inf Div) <->
  eval a v \/ evalinfrev a Div.
Proof.
intros.
split.
apply coevalprefinorcoevalpreinf_evalorevalinfrev.
apply evalorevalinfrev_coevalprefinorcoevalpreinf.  
Qed.



(** Well-behaved coevaluation extension **)
(* Step-indexed coevaluation *)

CoInductive coevalwb: nat -> term -> fdvalue -> Prop :=
| coevalwb_const: forall i, forall c,
    coevalwb i (Const c) (Fin (Const c))

| coevalwb_fun: forall i, forall x a,
    coevalwb i (Lam x a) (Fin (Lam x a))

| coevalwb_app: forall i, forall a1 x b,
    coevalwb i a1 (Fin (Lam x b)) ->
    forall a2 v2,
    coevalwb i a2 (Fin v2) ->
    forall v,
    coevalwb i (subst x v2 b) (Fin v) ->
    coevalwb (S i)  (App a1 a2) (Fin v)

| coevalwb_app_f: forall i, forall a1 x b,
    coevalwb i a1 (Fin (Lam x b)) ->
    forall j, forall a2 v2, coevalwb j a2 (Fin v2) ->
    forall k,
    coevalwb k (subst x v2 b) (Inf Div) ->
    coevalwb k (App a1 a2) (Inf Div)
           
| coevalwb_app_l: forall k, forall a1,
    coevalwb k a1 (Inf Div) ->
    forall a2,
    coevalwb k (App a1 a2) (Inf Div)

| coevalwb_app_r: forall i, forall a1 v,
    coevalwb i a1 (Fin v) ->
     forall k, forall a2,
    coevalwb k a2 (Inf Div) ->
    coevalwb k (App a1 a2) (Inf Div).


Require Import Lia.


Lemma coevalwbfin_monoticity:
forall i j,
  i <= j ->
forall a v,
  coevalwb i a (Fin v) ->
  coevalwb j a (Fin v).
Proof.
intros.
dependent induction i.

inversion_clear H0.
constructor.
constructor.

inversion_clear H0.
constructor.
constructor.

destruct j.
lia.
assert(i <= j).
lia.
econstructor 3.
apply IHi.
trivial.
exact H1.
apply IHi.
trivial.
exact H2.
apply IHi.
trivial.
exact H3.
Qed.


Lemma eval_coevalwbfin:
forall a v,
eval a v ->
exists i, coevalwb i a (Fin v).
Proof.
intros.
dependent induction H.
exists 0.
constructor.
exists 0.
constructor.
destruct IHeval1 as [i].
destruct IHeval2 as [j].
destruct IHeval3 as [k].
exists(S (max i (max j k))).
econstructor 3.
eapply coevalwbfin_monoticity with (i := i).
lia.
exact H2.
eapply coevalwbfin_monoticity with (i := j).
lia.
exact H3.
eapply coevalwbfin_monoticity with (i := k).
lia.
exact H4.
Qed.


Lemma coevalwbfin_eval:
forall a  v,
(exists i, coevalwb i a (Fin v)) ->
eval a v.
Proof.
intros.
destruct H as [i].
revert H.
revert a.
revert v.
revert i.
intros.
dependent induction i.
inversion_clear H.
constructor.
constructor.

inversion_clear H.
constructor.
constructor.

econstructor 3.
apply IHi.
exact H0.
apply IHi.
exact H1.
apply IHi.
exact H2.
Qed.


Lemma coevalwbfin_eval_alt:
forall i, forall a v,
coevalwb i a (Fin v) ->
eval a v.
Proof.
intros.
apply coevalwbfin_eval.
exists i.
trivial.
Qed.


Lemma eval_coevalwbfin_eq:
forall a v, eval a v <-> exists i, coevalwb i a (Fin v).
Proof.
intros.
split.
intros.
apply eval_coevalwbfin.
trivial.
intros.
destruct H as [i].
eapply coevalwbfin_eval.
exists i.
exact H.
Qed.


Lemma evalinfrev_coevalwbinf:
forall a,
  evalinfrev a Div ->
forall k,
  coevalwb k a (Inf Div).
Proof.
cofix CH.
intros.
inversion H.

subst.
constructor 5.
apply CH.
trivial.

subst.
apply eval_coevalwbfin in H0.
destruct H0 as [i].
econstructor 6.
exact H0.
apply CH.
exact H1.

subst.
apply eval_coevalwbfin in H0.
destruct H0 as [i].
apply eval_coevalwbfin in H1.
destruct H1 as [j].
econstructor 4.
exact H0.
exact H1.
apply CH.
exact H2.
Qed.


Lemma coevalwbinf_evalinfrev:
forall k, forall a,
coevalwb k a (Inf Div) ->
evalinfrev a Div.
Proof.
cofix CH.

intros.
inversion H.

subst.
econstructor 3.
apply coevalwbfin_eval.
exists i.
exact H0.
eapply coevalwbfin_eval.
exists j.
exact H1.
eapply CH.
exact H2.

subst.
constructor 1.
eapply CH.
exact H0.

subst.
econstructor 2.
eapply coevalwbfin_eval.
exists i.
exact H0.
eapply CH.
exact H1.
Qed.



Lemma evalinfrev_coevalwbinf_eq:
forall k, forall a, evalinfrev a Div <-> coevalwb k a (Inf Div).
Proof.
intros.
split.
intros.
apply evalinfrev_coevalwbinf.
trivial.
intros.
eapply coevalwbinf_evalinfrev.
exact H.
Qed.


(*Step-indexed coevaluation is equivalent to evaluation union divergence*)
Theorem coevalwb_eval_or_evalinfrev_eq:
forall k, forall a v,
(exists i, coevalwb i a (Fin v)) \/ coevalwb k a (Inf Div) <-> eval a v \/ evalinfrev a Div.
Proof.
intros.
assert(forall p q r s,  ((p <-> q) /\ (r <-> s)) -> ((p \/ r) <-> (q \/ s))).
firstorder.
apply H.
split.
symmetry.
apply eval_coevalwbfin_eq.
symmetry.
apply evalinfrev_coevalwbinf_eq.
Qed.


(*The finite part of step-indexed coevaluation is deterministic*)
Lemma coevalwbfin_deterministic:
  forall i, forall a v1,
    coevalwb i a (Fin v1) ->
  forall j, forall v2,
    coevalwb j a (Fin v2) ->        
    v1 = v2.
Proof.
intros.
apply coevalwbfin_eval_alt in H.
apply coevalwbfin_eval_alt in H0.
eapply eval_deterministic.
exact H.
exact H0.

Restart.

intros.
dependent induction i.
inversion H.
subst.
inversion_clear H0.
trivial.

subst.
inversion_clear H0.
trivial.

inversion H.
subst.
inversion_clear H0.
trivial.

subst.
inversion_clear H0.
trivial.

subst.
inversion_clear H0.
assert(Lam x b = Lam x0 b0).
eapply IHi.
exact H3.
exact H1.
assert(v0 = v3).
eapply IHi.
exact H4.
exact H2.
inversion H0.
subst.
eapply IHi.
exact H6.
exact H5.
Qed.


(*The infinite part of step-indexed coevaluation is deterministic
   in the sense that if a term coevaluates to the singleton 
   infinite value Div for some index, then it coevaluates to the 
   same singleton infinite value Div for any index*)
Lemma coevalwbinf_deterministic:
forall k, forall a, coevalwb k a (Inf Div) ->
forall l, coevalwb l a (Inf Div).
Proof.
cofix CH.

intros.
inversion H.
subst.
econstructor 4.
exact H0.
exact H1.
eapply CH.
exact H2.

subst.
constructor 5.
eapply CH.
exact H0.

subst.
econstructor 6.
exact H0.
eapply CH.
exact H1.
Qed.



(*The finite and infinite parts of step-indexed coevaluation are mutually exclusive*)
Lemma coevalwb_fin_inf_exclusive:
forall i, forall a  v,
  coevalwb i a (Fin v) ->
forall k, coevalwb k a (Inf Div) ->
  False.
Proof.
intros.
apply coevalwbfin_eval_alt in H.
apply coevalwbinf_evalinfrev in H0.
apply evalinf_evalinfrev_eq in H0.
eapply eval_evalinf_exclusive.
exact H.
exact H0.


Restart.
  
intros.
dependent induction i.
inversion H.
subst.

inversion H0.

subst.
inversion H0.

inversion H.

subst.
inversion H0.

subst.
inversion H0.

subst.

inversion H0.

subst.

assert(Lam x b = Lam x0 b0).
eapply coevalwbfin_deterministic.
exact H3.
exact H5.
inversion H1.

assert(v2 = v0).
eapply coevalwbfin_deterministic.
exact H4.
exact H8.
subst.
eapply IHi.
exact H6.
exact H9.

subst.
eapply IHi.
exact H3.
exact H5.

subst.
eapply IHi.
exact H4.
exact H8.
Qed.



(** Expected coevaluation behavior **)

(*Step-indexed coevaluation is equivalent to evaluation or divergence, as appropriate*)
Theorem coevalwb_eval_or_evalinf_eq:
forall a vg,
  (exists i, coevalwb i a vg) <->
  match vg with
  | (Fin v) => eval a v
  | (Inf Div) => evalinf a
  end.
Proof.
intros.  
destruct vg as [v|].
symmetry.
apply eval_coevalwbfin_eq.
destruct d.
assert(forall k, evalinfrev a Div <-> coevalwb k a (Inf Div)).
intros.
apply evalinfrev_coevalwbinf_eq.
symmetry in H.
split.
intros.
destruct H0 as [i].
specialize H with (k := i).
apply evalinf_evalinfrev_eq.
apply H.
trivial.
intros.
exists 0.
apply H.
apply evalinf_evalinfrev_eq.
trivial.
Qed.


(** Step-indexed coevaluation does not have known unusual features
      as illustrated by the following lemmas **)

Lemma coevalwb_omega_inf: forall i, coevalwb i omega (Inf Div).
Proof.
intros.
apply evalinfrev_coevalwbinf.
apply evalinf_evalinfrev_eq.
apply evalinf_omega.  

Restart.
cofix CH.
econstructor 4 with (i := 0).
constructor 2.
constructor 2 with (i := 0).
simpl.
fold delta.
fold omega.
apply CH.
Qed.

Lemma coevalwb_alpha_inf: forall i, coevalwb i alpha (Inf Div).
Proof.
intros.
apply evalinfrev_coevalwbinf.
apply evalinf_evalinfrev_eq.
apply evalinf_alpha.

Restart.

intros.
constructor 5.
apply coevalwb_omega_inf.
Qed.



Lemma not_coevalwb_omega_fin: forall i, forall v, not(coevalwb i omega (Fin v)).
Proof.
intros.
unfold not.
intros.
apply coevalwbfin_eval_alt in H.
apply not_eval_omega in H.
trivial.

Restart.

intros.
unfold not.
intros.
dependent induction i.
inversion H.

inversion_clear H.
inversion H0.
subst.
fold delta in H0.
inversion H1.
subst.
fold delta in H1.
fold delta in H2.
inversion H2.
subst.
eapply IHi.
unfold omega.
econstructor 3.
exact H6.
exact H7.
exact H8.

Restart.

intros.
assert(coevalwb i omega (Inf Div)).
apply coevalwb_omega_inf.
unfold not.
intros.
eapply coevalwb_fin_inf_exclusive.
exact H0.
exact H.
Qed.


Lemma coevalwb_beta_inf: forall i, coevalwb i beta (Inf Div).
Proof.
intros.
apply evalinfrev_coevalwbinf.
apply evalinf_evalinfrev.
apply evalinf_beta.

Restart.

intros.
unfold beta.

constructor 6 with (i := 0) (v := (Lam "x" (Const 0))).
constructor 2.
apply coevalwb_omega_inf.
Qed.

Lemma coevalwb_beta_only_inf:
forall i, forall vg,
  coevalwb i beta vg ->
  vg = (Inf Div).
Proof.
intros.
assert(coevalwb i beta (Inf Div)).
apply coevalwb_beta_inf.
destruct vg as [v|].
exfalso.
eapply coevalwb_fin_inf_exclusive.
exact H.
exact H0.
destruct d.
trivial.
Qed.
